/**
 * http://www.ideawu.net
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define ROUTE_UPDATE_INTERVAL 30
#define ROUTE_BROADCAST_INTERVAL 1
#define BUF_SIZE (16 * 128)
#define TTL 26

#define RECORD_LEN 16

#define   MAX_POINTS   26
#define   MAX_SIGNPOSTLENGTH   100

#define NODE_FAILURE_THRESH 3

struct ROUTE{
	int next[MAX_POINTS],length[MAX_POINTS];
	int ttl;
	int tick;
	int now_way;
	int neibor_count;
	int flag;
};

struct Path{
	int node_count;
	int nodes[MAX_SIGNPOSTLENGTH];
	int length;
};

// server config
struct Config{
	int id;
	int port;
	int base_port;
};

struct Neibors{
	int size;
	int id[MAX_POINTS];
	int port[MAX_POINTS];
};

struct ROUTE route[MAX_POINTS];
struct Path path;
struct Config config;
struct Neibors neibors;

int signpost[MAX_SIGNPOSTLENGTH];
int signpostlen;

int sock;
pthread_mutex_t route_info_lock;

void *broadcast_recv(void *argv);
void *broadcast_send(void *argv);


// dijkstra algorithm, modified from void ways(int now_point,int step)
void find_path(int src_id, int dst_id, int step){
	static int length = 0;
	int   next_id;
	int   i;

	length += step;

	if(src_id == dst_id){
		if(length<path.length){
			path.length = length;
			for(i = 0;i<signpostlen;i++){
				path.nodes[i] = signpost[i];
			}
			path.node_count = signpostlen;
		}
		length -= step;
		signpostlen --;
		return;
	}
	route[src_id].flag = 1;

	int n;
	for(n = 0; n < route[src_id].neibor_count; n++){
		next_id = route[src_id].next[n];
		if(!route[next_id].flag || next_id == dst_id){
			signpost[signpostlen++] = next_id;
			find_path(next_id, dst_id, route[src_id].length[n]);
		}
	}
	length -= step;
	signpostlen--;

	route[src_id].flag = 0;

	return;
}


int direct_linked(int src, int dst){
	int m;
	for(m=0; m<route[src].neibor_count; m++){
		if(route[src].next[m] == dst){
			return 1;
		}
	}
	return 0;
}

int is_my_neibor(int node){
	int m;
	for(m=0; m<neibors.size; m++){
		// if is neibor
		if(neibors.id[m] == node){
			return 1;
		}
	}
	return 0;
}


void broadcast(){
	socklen_t addr_len;
	struct sockaddr_in addr;
	pthread_t recv_t_id;
	pthread_t send_t_id;

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(config.port);

	if((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
		printf("Create socket failed.\n");
		return;
	}

	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0){
		printf("Bind to %d failed.\n", config.port);
		return;
	}else{
		//printf("Socket bind to port %d\n", config.port);
	}

	// start receiver thread
	if(pthread_create(&recv_t_id, NULL, broadcast_recv, NULL) != 0){
		printf("Can't create thread broadcast_recv.\n");
		exit(1);
	}

	// start sender thread
	if(pthread_create(&send_t_id, NULL, broadcast_send, NULL) != 0){
		printf("Can't create thread broadcast_send.\n");
		exit(1);
	}
}

/*
Packet format:
sender, (, src, dst, cost, ttl)+
4,(,4,4,4,4)+
*/

// run as a thread
void *broadcast_recv(void *argv){
	int r_size;
	char addr_p[32];
	int buf[BUF_SIZE];
	struct sockaddr_in addr;
	socklen_t addr_len = sizeof(struct sockaddr_in);

	while(1){
		r_size = recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr*)&addr, &addr_len);
		inet_ntop(AF_INET, &(addr.sin_addr.s_addr), addr_p, 32);
		if((r_size-4) % RECORD_LEN != 0){
			//printf("%s\n", strerror(errno));
			printf("Bad packet received from %s:%d. r_size = %d\n", addr_p, ntohs(addr.sin_port), r_size);
			continue;
		}
		//printf("Receive from %s:%d, records=%d\n", addr_p, ntohs(addr.sin_port), r_size/RECORD_LEN);

		int m, n;
		int src, dst, cost, ttl;
		int pos;
		int sender = ntohl(buf[0]);

	pthread_mutex_lock(&route_info_lock);
		for(pos=1; pos<r_size/4; pos+=4){
			src = ntohl(buf[pos]);
			dst = ntohl(buf[pos + 1]);
			cost = ntohl(buf[pos + 2]);
			ttl = ntohl(buf[pos + 3]);
			//printf("recv %c->%c=%d, ttl=%d\n", src+'A', dst+'A', cost, ttl);
						
			if(src == config.id && dst != sender || dst == config.id && src != sender){
				continue;
			}
			if(is_my_neibor(src) && src != sender || is_my_neibor(dst) && dst != sender){
				continue;
			}
			
			int is_new_record = 1;
			for(m=0; m<route[src].neibor_count; m++){
				if(route[src].next[m] == dst){
					is_new_record = 0;
					route[src].tick = 0;
					route[dst].tick = 0;
					if(ttl > route[src].ttl){
						route[src].ttl = ttl;
						route[dst].ttl = ttl;
						//printf("%c->%c=%d, update to %d\n", src+'A', dst+'A', route[src].length[m], cost);
						// update src->dst
						route[src].length[m] = cost;
						// update dst->src
						for(n=0; n<route[dst].neibor_count; n++){
							if(route[dst].next[n] == src){
								route[dst].length[n] = cost;
								break;
							}
						}
					}
					break;
				}
			}
			if(is_new_record){
				//printf("add new node %c->%c\n", src+'A', dst+'A');
				// add new record
				route[src].next[route[src].neibor_count] = dst;
				route[src].length[route[src].neibor_count] = cost;
				route[src].neibor_count++;
				route[src].ttl = ttl;

				route[dst].next[route[dst].neibor_count] = src;
				route[dst].length[route[dst].neibor_count] = cost;
				route[dst].neibor_count++;
				route[dst].ttl = ttl;
				route[src].tick = 0;
				route[dst].tick = 0;
			}
		}
		route[sender].ttl = TTL;
	pthread_mutex_unlock(&route_info_lock);
	}
}


void *broadcast_send(void *argv){
	int i, j, k;
	int buf[BUF_SIZE];
	char addr_p[32] = "127.0.0.1";
	int len = 0;
	int src, dst, cost, ttl;
	struct sockaddr_in addr;	
	socklen_t addr_len = sizeof(struct sockaddr_in);

	addr.sin_family = AF_INET;
	inet_pton(AF_INET, addr_p, &(addr.sin_addr.s_addr));

	while(1){
		sleep(ROUTE_BROADCAST_INTERVAL);
		buf[0] = htonl(config.id);
		len = 1;
	pthread_mutex_lock(&route_info_lock);
		// increase all known nodes' tick
		// and remove dead nodes.
		int m, n;
		for(m=0; m<MAX_POINTS; m++){
			if(m == config.id){
				continue;
			}
			route[m].tick ++;
			if(route[m].ttl <= 0 || route[m].tick > NODE_FAILURE_THRESH && is_my_neibor(m)){
				if(route[m].neibor_count > 0){
					//printf("remove node %c\n", m+'A');
				}
				// remove all routes that starts from m.
				route[m].neibor_count = 0;
				// then remove all routes that point to m.
				for(n=0; n<MAX_POINTS; n++){
					for(i=0; i<route[n].neibor_count; i++){
						if(route[n].next[i] == m){
							route[n].neibor_count --;
							while(i < route[n].neibor_count){
								// all nodes behind m move forward one step
								route[n].next[i] = route[n].next[i+1];
								route[n].length[i] = route[n].length[i+1];
								i++;
							}
							break;
						}
					}
				}
			}
		}
		
		for(i=0; i<MAX_POINTS; i++){
			route[i].ttl --;
			ttl = htonl(route[i].ttl);
			for(j=0; j<route[i].neibor_count; j++){
				if(i >= route[i].next[j]){
					continue;
				}
				src = htonl(i);
				dst = htonl(route[i].next[j]);
				cost = htonl(route[i].length[j]);
				buf[len] = src;
				buf[len + 1] = dst;
				buf[len + 2] = cost;
				buf[len + 3] = ttl;
				len += 4;
			}
		}
		
		// send to neibors
		for(i=0; i<route[config.id].neibor_count; i++){
			addr.sin_port = htons(config.base_port + route[config.id].next[i]);
			sendto(sock, buf, len*4, 0, (struct sockaddr*)&addr, addr_len);
			//printf("Send to %s:%d.\n", addr_p, ntohs(addr.sin_port));
		}
	pthread_mutex_unlock(&route_info_lock);
	}
}


int main(int argc, char **argv)
{	
	int start_point;
	int end_point;
	int   i,j,k;
	char dst_name;
	int cost, port;
	FILE   *fp;
	char filename[128] = "config_.txt";

	if(argc < 2){
		printf("Usage: %s [id]\n", argv[0]);
		printf("\tid = [A-Z]\n");
		return 1;
	}
	if(argv[1][0] < 'A' || argv[1][0] > 'Z'){
		printf("Invalid argument! id=[A-Z]\n");
		return 1;
	}
	config.id = argv[1][0] - 'A';
	config.base_port = 20000;
	config.port = config.id + config.base_port;

	filename[strlen("config")] = argv[1][0];
	pthread_mutex_init(&route_info_lock, NULL);

	for(i = 0;i<MAX_POINTS;i++){
		route[i].neibor_count = 0;
		route[i].flag = 0;
		route[i].tick = 0;
	}

	char line[128];
	char *buf;
	char *arg;
	int m = 0;
	fp = fopen(filename,"r");
	fgets(line, 128, fp);
	int line_count = atoi(line);
	while(m < line_count){
		m ++;
		buf = fgets(line, 128, fp);
		if(buf == NULL){
			break;
		}
		arg = strtok(buf, " ");
		if(arg != NULL){
			dst_name = arg[0];
		}else{
			printf("Bad config file: %s\n", filename);
			exit(1);
		}
		arg = strtok(NULL, " ");
		if(arg != NULL){
			cost = atoi(arg);
		}else{
			printf("Bad config file: %s\n", filename);
			exit(1);
		}
		arg = strtok(NULL, " ");
		if(arg != NULL){
			port = atoi(arg);
		}else{
			printf("Bad config file: %s\n", filename);
			exit(1);
		}

		i = config.id;
		j = dst_name - 'A';
		k = cost;
		
		neibors.id[neibors.size] = j;
		neibors.port[neibors.size] = port;
		neibors.size ++;

		route[i].next[route[i].neibor_count] = j;
		route[j].next[route[j].neibor_count] = i;
		route[i].length[route[i].neibor_count] = k;
		route[j].length[route[j].neibor_count] = k;
		route[i].neibor_count++;
		route[j].neibor_count++;
		route[i].ttl = TTL;
		route[j].ttl = TTL;
	}
	fclose(fp);

	broadcast();

	while(1){
		sleep(ROUTE_UPDATE_INTERVAL);

		int src, dst;
		src = config.id;

	pthread_mutex_lock(&route_info_lock);
		/*
		for(i=0; i<MAX_POINTS; i++){
			if(route[i].neibor_count == 0){
				continue;
			}
			for(j=0; j<route[i].neibor_count; j++){
				//if(i < route[i].next[j]){
					printf("%c->%c=%d, ttl=%d;", i+'A', route[i].next[j]+'A', route[i].length[j], route[i].ttl);
				//}
			}
			printf("\n");
		}
		*/
		for(dst=0; dst<MAX_POINTS; dst++){
			// reset global variables
			signpostlen = 1;
			path.length = 10000; 
			path.node_count = 1; 

			signpost[0] = src;

			find_path(src, dst, 0);
			
			if(path.node_count == 1){
				continue;
			}
			printf("least-cost path to node %c is: ", dst + 'A');
			for(i = 0; i<path.node_count; i++){
				printf("%c", path.nodes[i] + 'A');
			}
			printf(", and the cost is %.1f;\n", (float)path.length);
		}
		printf("\n");
	pthread_mutex_unlock(&route_info_lock);
	}
	return 0;
}
